#version 150

precision highp float;
precision highp int;

out vec4 FragColor;
in vec2 TexCoord;
uniform sampler2D 		FrontPositionMap;
uniform sampler2D 		BackPositionMap;
uniform sampler2D 		NormalMap;

struct AOControlSt {
	int 	MapWidth;
	int 	MapHeight;
	float 	SampleSize;
	float 	FallOff;
	float	Quality;
	float 	Contrast;
	float 	DistanceScale;
	int		Noise;
	int		UseBack;
};

uniform AOControlSt AOControl;

uniform mat4 renderPJInverseMatrix;	//	used to convert back to eye space coordinate

float rand(vec2 co){
    return fract(sin(dot(co ,vec2(12.9898,78.233))) * 43758.5453);
}

vec3 computeFrontPos(vec2 uv)
{
	vec4 pos = (renderPJInverseMatrix*vec4(2.0*uv-1.0,2.0*texture(FrontPositionMap, uv).r-1.0,1.0));	//back to eye space
	return pos.xyz/pos.w;
}
vec3 computeBackPos(vec2 uv)
{
	vec4 pos = (renderPJInverseMatrix*vec4(2.0*uv-1.0,2.0*texture(BackPositionMap, uv).r-1.0,1.0));	//back to eye space
	return pos.xyz/pos.w;
}

float computeAOPx(vec2 uv, vec3 pos, vec3 N, float bias, float d_scale, inout float sum_pts)
{
	vec3 tex_frontpos = computeFrontPos(uv);	//texture(FrontPositionMap, uv);
	//if (tex_frontpos.a>0.0001)	/*MJY 06/19/14*/
	{
		sum_pts = sum_pts + 1.0;
		vec3 diff = tex_frontpos.xyz - pos;
		vec3 v = normalize(diff);
		float d = length(diff)*d_scale;
		return max(0.0,(dot(N,v)-bias)*(1.0/(1.0+d)));
	}
	//else return 0.0;
}

float computeAOPxBackface(vec2 uv, vec3 pos, vec3 N, float bias, float d_scale, inout float sum_pts)
{
	vec3 tex_backpos = computeBackPos(uv);	//texture(BackPositionMap, uv);
	//	=if (tex_backpos.a>0.0001)	/*MJY 06/19/14*/
	{
		sum_pts = sum_pts + 1.0;
		vec3 diff = tex_backpos.xyz - pos;
		vec3 v = normalize(diff);
		float d = length(diff)*d_scale;
		return max(0.0,(dot(N,v)-bias)*(1.0/(1.0+d)));
	}
	//else return 0.0;
}

float computeAO(vec2 uv0, vec3 frontpos, vec3 normal, float ao_sampleRadius, int ao_numSamples, int ao_directions, float ao_dScale, float ao_bias, float ao_noise)
{
	float ao = 0.0;
	float randf = 0.0;
	if (ao_noise>0.0) randf = rand(uv0)*2.0-1.0;
	float num_pts = 0.0;
	for (int i=0;i<ao_directions;i++)
	{
		// get current direction
		float angle = ((float(i)+randf*0.25)/float(ao_directions))*2.0*3.14159;
		vec2 coord = vec2(cos(angle),sin(angle))*ao_sampleRadius;
		// each direction
		for (int j=0;j<ao_numSamples;j++)
		{
			ao += computeAOPx(uv0 + coord*(float(j)+1.0+randf*0.25)/float(ao_numSamples), frontpos, normal, ao_bias, ao_dScale, num_pts);
			if (AOControl.UseBack>0) ao += computeAOPxBackface(uv0 + coord*(float(j)+1.0+randf*0.25)/float(ao_numSamples), frontpos, normal, ao_bias, ao_dScale, num_pts);
		}
	}
	return ao/num_pts;
}
	
	
void main (void)
{
	float ao = 0.0;
	vec2 uv_this = TexCoord;
	vec3 tex_frontpos = computeFrontPos(uv_this);	//texture(FrontPositionMap, uv_this);
	//if (tex_frontpos.a>0.0001)		// valid fragment	/*MJY 06/19/14*/
	{
		vec3 tex_normal = texture(NormalMap, uv_this).xyz;
		// get user-specified parameters for AO
        float z_scale = 1.0/max((-tex_frontpos.z), 0.1);             		/*MJY 08/08/13 : avoid too small divisor*/
		float SampleSizef = min(AOControl.SampleSize*z_scale, 0.05);
		// int numDirections = int((1-depth*depth)*AO_Quality*max(AO_SampleSize,0)+6);
		// int numSamples = int((1-depth*depth)*AO_Quality*max(AO_SampleSize,0)+6); 
		float SampleSizeOnScreen = SampleSizef*float(AOControl.MapWidth);
		//int numDirections = int(AOControl.Quality*max(SampleSizeOnScreen/8,0)+4);
		//int numSamples = int(AOControl.Quality*max(SampleSizeOnScreen/3-6,0)+6); 
		
		/*MJY 06/13/14 : use a proper number of samples to avoid crash when using massive ao map*/
		int numDirections = int(AOControl.Quality*8.0+4.0);
		int numSamples = int(min(8.0, AOControl.Quality*SampleSizeOnScreen/10.0+2.0));
		
		
		float dScaleMax = max(1000.0/(AOControl.DistanceScale*AOControl.SampleSize), 0.01);
		float dScaleMin = 0.0;
		float dScale = AOControl.FallOff*(dScaleMax-dScaleMin)+dScaleMin;
		ao = computeAO(uv_this, tex_frontpos.xyz, tex_normal, SampleSizef, numSamples, numDirections, dScale, 0.05, float(AOControl.Noise));
		float ContrastMax = max(dScaleMax*AOControl.SampleSize,1.0);
		float ContrastMin = 0.0;
		float ao_contrast = AOControl.Contrast*(ContrastMax-ContrastMin)+ContrastMin;
		ao = clamp(ao * ao_contrast,0.0,1.0);	// intensity
	}

	FragColor = vec4(1.0-ao,1.0-ao,1.0-ao,1.0);
}
