#version 150

precision highp float;
precision highp int;

#define LIGHT_TYPE_AMBIENT		0
#define LIGHT_TYPE_POINT		1
#define LIGHT_TYPE_DISTANT		2
#define	LIGHT_TYPE_CONE			3
#define ALPHA_THRES             0.009   /*above which the pixel is rendered*/
#define DO_PROC_TEX             0

#define PI 3.141592653

// final color output
out vec4 FragColor;

// input var from vertex shader
in vec3 MCVertexToFrag;
in vec3 NormalToFrag;
in vec4 PxPosToFrag;
in vec4 ProjPxPosToFrag;
//in vec4 LightPosToFrag;
//in vec4 LightDirToFrag;
in vec4 PxPosinLightsView;
in vec4 ShadowMapTexCoord;
in float depthLinear;
in vec2 TexCoordToFrag;
in vec4 TangentToFrag;
in vec4 ColorToFrag;

// Texture samplers
uniform sampler2D 		ColorTexture;
uniform sampler2D 		BumpMap;
uniform sampler2D 		Shadowmap2D;
uniform samplerCube		ShadowmapCube;
uniform sampler2D 		AmbientOcclusionMap;

// light
struct LightSt{
	int 	Type;
	// generic
	vec3 	Color;
	float	Factor;
	int 	Falloff;
	int		ShadowOn;
	// location
	vec3 	Origin;
	vec3 	Direction;
	// point light
	float 	Radius;
	// cone light
	float	InAngle;
	float	OutAngle;
	// view
	float 	NearClip;
	float 	FarClip;
	// control
	int		InEyeSpace;
} ;

uniform LightSt Light;

uniform int Light_Num;

struct MaterialSt{
	vec3 	Ambient;
	vec3 	Diffuse;
	vec3 	Specular;
	float	Shininess;
	int 	Has_Specular;
	int 	Has_Texture;
    int     Do_Sphere_Map;
    int     Has_Vertex_Color;
	int 	Has_Bump;
    int     Receive_Light;
	int 	Is_Transparent;
    float   Transparency;
	int 	Receive_Shadow;
	int		Cast_Shadow;
	int		Use_AO;
    
    int     Const_Refl;     /*constant reflection on/off*/
} ;

uniform MaterialSt Material;

struct ShadowControlSt {
    int 	MapWidth;
	int 	MapHeight;
	int 	CubeSize;
	int 	CubeFace;
	float 	SmoothQuality;
	float 	Softness;
	float   Darkness;
	int     ShadowOnGlobal;
};

uniform ShadowControlSt ShadowControl;

struct AOControlSt{
	int	  IsEnabled;
} ;

uniform AOControlSt AOControl;


float reduceBackfaceTrans = 0.0;    // for transparent objects, we suppose that a portion of the light reached the back face(from the view of the light)
float ambientInWorkMode = 0.3;      // ambient light in work mode, TODO: make this value user-controllable
//float bias = 0.005;               //  bias to avoid self shadowing

// point light illumination
// MJY 08/07/13 TODO : more attenuation options
void PointLight(in vec3 eye, in vec3 pxPos, in vec3 lightPos, in vec3 N,
				inout vec3 ambient, inout vec3 diffuse, inout vec3 specular)
{
	vec3 L = lightPos - pxPos;
	//float d2 = dot(L,L);
    //float attenuation = 1.0/(1.0+0.00000001*d2);	// TODO falloff?
    float d2 = length(L);
    float attenuation = 1.0/(1.0+d2/Light.Radius);
    
	L = normalize(L);
    
    float dotN = dot(L,N);
    float LdotN;
    
    /* MJY 06/12/13, if transparent, the back facess of the object receive some light*/
    if (Material.Is_Transparent > 0)
        LdotN = max(-dotN*reduceBackfaceTrans, dotN);
    else
        LdotN = max(0.0,dotN);
        
	diffuse += Light.Color * LdotN *attenuation;
	
	if (Material.Has_Specular>0 &&  dotN >= 0.0001)
	{
		vec3 R = reflect(L,N);
		vec3 V = normalize(pxPos - eye);
		float VdotR = max(0.0,dot(V,R));
		float pf = 0.0;
		if(Material.Shininess>=0.0001)
			pf = pow(VdotR, Material.Shininess);
			
		specular += Light.Color * pf *attenuation;
	}
	// TODO ambient light?
	// ambient
}

// directional light illumination
void DirectionalLight(in vec3 eye, in vec3 pxPos, in vec3 lightDir, in vec3 N,
				inout vec3 ambient, inout vec3 diffuse, inout vec3 specular)
{
	vec3 L = -lightDir;

    float dotN = dot(L,N);
    float LdotN;
    
    /* MJY 06/12/13, if transparent, the back facess of the object receive some light*/
    if (Material.Is_Transparent > 0)
    {
        LdotN = max((-dotN)*reduceBackfaceTrans, dotN);
    }
    else    
        LdotN = max(0.0,dotN);
    
	diffuse += Light.Color * LdotN;
	
	if (Material.Has_Specular>0 && dotN >= 0.0001)
	{
		vec3 R = reflect(L,N);
		vec3 V = normalize(pxPos - eye);
		float VdotR = max(0.0,dot(V,R));
		float pf = 0.0;
		if(Material.Shininess>=0.0001)
			pf = pow(VdotR, Material.Shininess);
			
		specular += Light.Color * pf;
	}

	// ambient is used only for eye-space light (work mode)
	if (Light.InEyeSpace>0) ambient += ambientInWorkMode * Light.Color;
}

// cone light illumination
void ConeLight(in vec3 eye, in vec3 pxPos, in vec3 lightPos, in vec3 lightDir, in vec3 N,
				inout vec3 ambient, inout vec3 diffuse, inout vec3 specular)
{
	vec3 L = lightPos - pxPos;
	float d2 = dot(L,L);
	float attenuation = 1.0;//1.0/(1.0+0.0001*d2);	// TODO falloff?
	
	L = normalize(L);
	
	float LdotDir = -dot(L,lightDir);
	
	float litCone = -1.0;
	float cosin = cos(Light.InAngle/2.0);
	float cosout = cos(Light.OutAngle/2.0);
	if (LdotDir>cosin)	// fully lit
	{
		litCone = 1.0;
	}
	else if (LdotDir>cosout)	// gradually dim
	{
		litCone = (LdotDir - cosout)/(cosin - cosout);
		litCone = litCone * litCone;
	}
	if (litCone>0.0)
	{
        float dotN = dot(L,N);
        float LdotN;
        
        /* MJY 06/12/13, if transparent, the back face of the object is visible*/
        if (Material.Is_Transparent > 0)
            LdotN = max(-dotN*reduceBackfaceTrans, dotN);
        else
            LdotN = max(0.0,dotN);

		diffuse += Light.Color * LdotN *attenuation * litCone;
	
		if (Material.Has_Specular>0 && dotN >= 0.0001)
		{
			vec3 R = reflect(L,N);
			vec3 V = normalize(pxPos - eye);
			float VdotR = max(0.0,dot(V,R));
			float pf = 0.0;
			if( Material.Shininess>=0.0001)
				pf = pow(VdotR, Material.Shininess);
			
			specular += Light.Color * pf *attenuation * litCone;
		}
	}

	// TODO ambient light?
	// ambient
}

//	MJY 06/10/14: Variance Shadow Map, new functions
float linstep(float min_, float max_, float v)  
{  
  return clamp((v-min_) / (max_-min_), 0.0, 1.0);
}  
float ReduceLightBleeding(float p_max, float Amount)  
{  
  // Remove the [0, Amount] tail and linearly rescale (Amount, 1].  
   return linstep(Amount, 1.0, p_max);
} 
float ChebyshevUpperBound(vec2 Moments, float currentDepth)  
{  
	// One-tailed inequality valid if currentDepth > Moments.x  
	float p=0.0;
	
	//currentDepth could be out of [0,1] when it casts no light as we only consider casting-light objs when rendering the shadow map.
	//cleared to 1.2, Moments.x>1.1 means this depth buffer was never updated (no object blocked the light)
	if (currentDepth <= Moments.x || Moments.x>1.1)	
		p=1.0;
		
	// compute variance
	float Variance=Moments.y-(Moments.x*Moments.x);
	Variance = max(Variance, 0.00001);  
	// Compute probabilistic upper bound.
	float d=currentDepth-Moments.x;
	float p_max = Variance / (Variance + d*d);
	p_max=ReduceLightBleeding(p_max, 0.2);
	float ret = max(p, p_max);
	return ret;
}  

float ComputeShadow_Cubemap(vec3 shadowTexCoord, float currentDepth)
{
	float lit=1.0;
    
    float minFilterSize = 0.0;
    float maxFilterSize = min(float(ShadowControl.MapWidth/256), 20.0);
    float FilterSize = minFilterSize + ShadowControl.Softness*(maxFilterSize - minFilterSize);
    int filterSizei = int(FilterSize);
    float PixelSize = 1.0/float(ShadowControl.CubeSize);
    float minJumpSize = 1.0;
    float maxJumpSize = max(FilterSize/2.0,1.0);
    //float jump = (1.0-ShadowControl.SmoothQuality/2.0);
    int jump = int(minJumpSize + (1.0-ShadowControl.SmoothQuality)*(maxJumpSize-minJumpSize));
    
    vec3 maskI = vec3(shadowTexCoord.y, -shadowTexCoord.x, 0);
    vec3 maskJ = cross(shadowTexCoord, maskI);                  /*MJY TODO : find a better mask*/
    
	vec2 sumc = vec2(0.0, 0.0);   // intensity sampling accumulation
    float sumi = 0.0;   // number of sampling
	
	//	TODO: only blur Penumbra area, which is hard to determine
	for (int i=-filterSizei;i<=filterSizei;i=i+jump)
	{
		for (int j=-filterSizei;j<=filterSizei;j=j+jump)
		{
				vec3 coord=shadowTexCoord+float(i)*maskI*PixelSize+float(j)*maskJ*PixelSize;
				sumc+=texture(ShadowmapCube, coord).xy;			
				sumi+=1.0;		
		}
	}
	if (sumi>0.5)   // never fail logically
		sumc = sumc / sumi;
		
		
	lit=1.0-ShadowControl.Darkness*(1.0-ChebyshevUpperBound(sumc, currentDepth));

    return lit;
}

float ComputeShadow_2D(vec2 shadowTexCoord, float currentDepth)
{
	float lit=1.0;
    float minFilterSize = 0.0;
    float maxFilterSize = min(float(ShadowControl.MapWidth/256), 20.0);
    float FilterSize = minFilterSize + ShadowControl.Softness*(maxFilterSize - minFilterSize);
    int filterSizei = int(FilterSize);
    float PixelSizeX = 1.0/float(ShadowControl.MapWidth);
    float PixelSizeY = 1.0/float(ShadowControl.MapHeight);
    
    
    float minJumpSize = 1.0;
    float maxJumpSize = max(FilterSize/2.0,1.0);
    int jump = int(minJumpSize + (1.0-ShadowControl.SmoothQuality)*(maxJumpSize-minJumpSize));
    //float jump = (1.0-ShadowControl.SmoothQuality/2.0);
    
    vec2 sumc = vec2(0.0, 0.0);   // intensity sampling accumulation
    float sumi = 0.0;   // number of sampling
    
	//	TODO: only blur Penumbra area, which is hard to determine
	for (int i=-filterSizei;i<=filterSizei;i=i+jump)
	{
		for (int j=-filterSizei;j<=filterSizei;j=j+jump)
		{
				vec2 coord=shadowTexCoord+vec2(float(i)*PixelSizeX,float(j)*PixelSizeY);
				if (coord.x>=0.0&&coord.x<=1.0&&coord.y>=0.0&&coord.y<=1.0)
				{
					sumc+=texture(Shadowmap2D, coord).xy;			
					sumi+=1.0;
				}
				else
				{
					sumc+=vec2(1.2, 1.2*1.2);	// depth out of bounding box is 1.2		
					sumi+=1.0;			
				}			
		}
	}
	if (sumi>0.5)   // never fail logically
		sumc = sumc / sumi;
		
		
	lit=1.0-ShadowControl.Darkness*(1.0-ChebyshevUpperBound(sumc, currentDepth));
	
	return lit;
}

#if DO_PROC_TEX
vec3 mod289(vec3 x) {
  return x - floor(x * (1.0 / 289.0)) * 289.0;
}

vec4 mod289(vec4 x) {
  return x - floor(x * (1.0 / 289.0)) * 289.0;
}

vec4 permute(vec4 x) {
     return mod289(((x*34.0)+1.0)*x);
}

vec4 taylorInvSqrt(vec4 r)
{
  return 1.79284291400159 - 0.85373472095314 * r;
}

float snoise(vec3 v)
  { 
  const vec2  C = vec2(1.0/6.0, 1.0/3.0) ;
  const vec4  D = vec4(0.0, 0.5, 1.0, 2.0);

// First corner
  vec3 i  = floor(v + dot(v, C.yyy) );
  vec3 x0 =   v - i + dot(i, C.xxx) ;

// Other corners
  vec3 g = step(x0.yzx, x0.xyz);
  vec3 l = 1.0 - g;
  vec3 i1 = min( g.xyz, l.zxy );
  vec3 i2 = max( g.xyz, l.zxy );

  //   x0 = x0 - 0.0 + 0.0 * C.xxx;
  //   x1 = x0 - i1  + 1.0 * C.xxx;
  //   x2 = x0 - i2  + 2.0 * C.xxx;
  //   x3 = x0 - 1.0 + 3.0 * C.xxx;
  vec3 x1 = x0 - i1 + C.xxx;
  vec3 x2 = x0 - i2 + C.yyy; // 2.0*C.x = 1/3 = C.y
  vec3 x3 = x0 - D.yyy;      // -1.0+3.0*C.x = -0.5 = -D.y

// Permutations
  i = mod289(i); 
  vec4 p = permute( permute( permute( 
             i.z + vec4(0.0, i1.z, i2.z, 1.0 ))
           + i.y + vec4(0.0, i1.y, i2.y, 1.0 )) 
           + i.x + vec4(0.0, i1.x, i2.x, 1.0 ));

// Gradients: 7x7 points over a square, mapped onto an octahedron.
// The ring size 17*17 = 289 is close to a multiple of 49 (49*6 = 294)
  float n_ = 0.142857142857; // 1.0/7.0
  vec3  ns = n_ * D.wyz - D.xzx;

  vec4 j = p - 49.0 * floor(p * ns.z * ns.z);  //  mod(p,7*7)

  vec4 x_ = floor(j * ns.z);
  vec4 y_ = floor(j - 7.0 * x_ );    // mod(j,N)

  vec4 x = x_ *ns.x + ns.yyyy;
  vec4 y = y_ *ns.x + ns.yyyy;
  vec4 h = 1.0 - abs(x) - abs(y);

  vec4 b0 = vec4( x.xy, y.xy );
  vec4 b1 = vec4( x.zw, y.zw );

  //vec4 s0 = vec4(lessThan(b0,0.0))*2.0 - 1.0;
  //vec4 s1 = vec4(lessThan(b1,0.0))*2.0 - 1.0;
  vec4 s0 = floor(b0)*2.0 + 1.0;
  vec4 s1 = floor(b1)*2.0 + 1.0;
  vec4 sh = -step(h, vec4(0.0));

  vec4 a0 = b0.xzyw + s0.xzyw*sh.xxyy ;
  vec4 a1 = b1.xzyw + s1.xzyw*sh.zzww ;

  vec3 p0 = vec3(a0.xy,h.x);
  vec3 p1 = vec3(a0.zw,h.y);
  vec3 p2 = vec3(a1.xy,h.z);
  vec3 p3 = vec3(a1.zw,h.w);

//Normalise gradients
  vec4 norm = taylorInvSqrt(vec4(dot(p0,p0), dot(p1,p1), dot(p2, p2), dot(p3,p3)));
  p0 *= norm.x;
  p1 *= norm.y;
  p2 *= norm.z;
  p3 *= norm.w;

// Mix final noise value
  vec4 m = max(0.6 - vec4(dot(x0,x0), dot(x1,x1), dot(x2,x2), dot(x3,x3)), 0.0);
  m = m * m;
  return 42.0 * dot( m*m, vec4( dot(p0,x0), dot(p1,x1), 
                                dot(p2,x2), dot(p3,x3) ) );
  }

 vec3 granite(vec3 pos)
 {
 	const vec3 Color1 = vec3(0.35, 0.3, 0.2);
	const vec3 Color2 = vec3(0.7, 0.7, 0.7);
	
	vec4 noisevec;
	noisevec.x = snoise(pos*1.0) * 8.0;
	noisevec.y = snoise(pos*2.0) * 4.0;
	noisevec.z = snoise(pos*4.0) * 2.0;
	noisevec.w = snoise(pos*8.0) * 1.0;
	
	noisevec = noisevec / 8.0;

	// alternate granite
	// float intensity = min(1.0, noisevec[2] * 12.0 + noisevec[3]*12);
	float intensity = min(1.0, noisevec[3] * 12.0);

	// vec3 color = vec3(intensity * LightIntensity);
	return mix(Color1, Color2, intensity);
 }
 
 vec3 wood(vec3 pos)
 {
 	const vec3 LightWood = vec3(0.6, 0.3, 0.1);
	const vec3 DarkWood = vec3(0.4, 0.2, 0.07);
	const vec3 NoiseScale = vec3(0.5, 0.1, 0.1);
	const float Noisiness = 5.0;
	const float RingFreq = 3.5;
	const float LightGrains = 1.0;
	const float DarkGrains = 0.0;
	const float GrainThreshold = 0.3;

	vec4 noisevec;
	noisevec.x = snoise(pos * NoiseScale*1.0) * 8.0 * Noisiness;
	noisevec.y = snoise(pos * NoiseScale*2.0) * 4.0 * Noisiness;
	noisevec.z = snoise(pos * NoiseScale*4.0) * 2.0 * Noisiness;
	noisevec.w = snoise(pos * NoiseScale*8.0) * 1.0 * Noisiness;
	// noisevec = (noisevec / 8.0 + 1.0) / 2.0;
	noisevec = noisevec / 8.0;

	vec3 location = pos + noisevec.xyz;
	float dist = sqrt( location.x * location.x + location.y * location.y ) * RingFreq;

	float r = fract(dist + noisevec[0] + noisevec[1] + noisevec[2]) * 2.0;

	if (r > 1.0)
		r = 2.0 - r;

	vec3 color = mix(LightWood, DarkWood, r);

	noisevec[2] *= r;

	if (r<GrainThreshold)
		color += LightWood * LightGrains * noisevec[2];
	else
		color += DarkWood * DarkGrains * noisevec[2];

	return color;
 }
 
 vec3 marble(vec3 pos)
 {
 	const vec3 Color1 = vec3(0.8, 0.8, 0.8);
	const vec3 Color2 = vec3(0.2, 0.2, 0.5);


	vec4 noisevec;
	noisevec.x = snoise(pos*1.0) * 8.0;
	noisevec.y = snoise(pos*2.0) * 4.0;
	noisevec.z = snoise(pos*4.0) * 2.0;
	noisevec.w = snoise(pos*8.0) * 1.0;
	
	noisevec = noisevec / 8.0;

	float intensity = abs(noisevec[0] - 0.20) + 
	                  abs(noisevec[1] - 0.10) + 
					  abs(noisevec[2] - 0.05) +
					  abs(noisevec[3] - 0.025);

	// intensity = intensity *  intensity;
	// intensity -= 0.3;

	float sineval = sin(pos.y * 12.0 + intensity * 8.0) * 0.5 + 0.5;

	return mix(Color1, Color2, sineval);
 }
 
  vec3 cloud(vec3 pos)
  {
 	const vec3 SkyColor = vec3(0.1, 0.2, 0.8);
	const vec3 CloudColor = vec3(0.8, 0.8, 0.8);
	const vec3 CloudColorDark = vec3(0.6, 0.6, 0.6);

	float noise = snoise(pos) * 8.0 + snoise(pos*2.0) * 4.0 + snoise(pos*4.0) * 2.0 + snoise(pos * 8.0);
	noise = (noise / 8.0 + 1.0) / 2.0;
	// noise = noise * noise;
	noise -= 0.4;
	noise *= 2.0;
	// noise = noise * noise;
	float intensity = noise;
	vec3 color = vec3(0.0, 0.0, 0.0);

	intensity = clamp(intensity, 0.0, 1.0);

	if (intensity < 0.5)
		color = mix(SkyColor, CloudColor, intensity*2.0);
	else
		color = mix(CloudColor, CloudColorDark, (intensity - 0.5)*2.0);

	return color; 
  }
  
  vec3 strip(vec3 pos)
  {
 	const float Frequency = 3.0;
	float V = pos.x+pos.y+pos.z;
	float sawtooth = fract(V * Frequency);
	float triangle = abs(2.0 * sawtooth - 1.0);
	float dp = length(vec2(dFdx(V), dFdy(V)));
	float edge = dp * Frequency * 2.0;
	float square = smoothstep(0.5 - edge, 0.5 + edge, triangle);

	vec3 Color1 = vec3(0.8, 0.8, 0.8);
	vec3 Color2 = vec3(0.1, 0.1, 0.15);
	return mix(Color1, Color2, square); 
  }
  
  vec3 ribbon(vec3 pos)
  {
  	const vec3 SkinColor1 = vec3(0.9, 0.7, 0.2);
	const vec3 SkinColor2 = vec3(0.2, 0.5, 0.2);

	float noise = snoise(pos) * 8.0 + snoise(pos*2.0) * 4.0 + snoise(pos*4.0) * 2.0 + snoise(pos * 8.0);
	noise = (noise / 8.0 + 1.0) / 2.0;
	// noise = noise * noise;
	 noise -= 0.3;
	 noise *= 2.0;
	// noise = noise * noise;
	float intensity = fract(noise * 3.14159265);
	vec3 color = vec3(0.0, 0.0, 0.0);

	intensity = clamp(intensity, 0.0, 1.0);
	if ( abs(intensity - 0.5) > 0.1 )
		color = SkinColor2;
	else
		color = SkinColor1;

	return color;
  }
  
  vec3 alien(vec3 pos)
  {
	const vec3 SkinColor1 = vec3(0.3, 0.5, 0.2);
	const vec3 SkinColor2 = vec3(0.8, 0.8, 0.5);
	const vec3 SkinColor3 = vec3(0.7, 0.8, 0.8);

	float noise = snoise(pos) * 8.0 + snoise(pos*2.0) * 4.0 + snoise(pos*4.0) * 2.0 + snoise(pos * 8.0);
	noise = (noise / 8.0 + 1.0) / 2.0;
	// noise = noise * noise;
	noise -= 0.3;
	noise *= 2.0;
	// noise = noise * noise;
	float intensity = noise;
	vec3 color = vec3(0.0, 0.0, 0.0);

	intensity = clamp(intensity, 0.0, 1.0);

	if (intensity < 0.50)
		color = mix(SkinColor1, SkinColor2, intensity/0.50);
	else
		color = mix(SkinColor1, SkinColor3, (intensity - 0.50)/0.50);

	return color;
  }
  
  vec3 turbulence(vec3 pos)
  {
  	const vec3 Color1 = vec3(0.8, 0.7, 0.0);
	const vec3 Color2 = vec3(0.6, 0.1, 0.0);
	
	vec4 noisevec;
	noisevec.x = snoise(pos*1.0) * 8.0;
	noisevec.y = snoise(pos*2.0) * 4.0;
	noisevec.z = snoise(pos*4.0) * 2.0;
	noisevec.w = snoise(pos*8.0) * 1.0;
	// noisevec = (noisevec / 8.0 + 1.0) / 2.0;
	noisevec = noisevec / 8.0;
	// noisevec = noisevec * noisevec;

	float intensity = abs(noisevec[0] - 0.20) + 
	                  abs(noisevec[1] - 0.10) + 
					  abs(noisevec[2] - 0.05) +
					  abs(noisevec[3] - 0.025);

	// intensity = intensity *  intensity;
	intensity -= 0.3;

	// intensity = clamp(intensity * 1.5, 0.0, 1.0);

	return mix(Color1, Color2, intensity);
  }
  
vec3 checker(vec3 pos)
{
	return mod(floor(pos.x) + floor(pos.y) + floor(pos.z), 2) == 0 ? 
					vec3(0, 0, 0) : vec3(1, 1, 1);
}

 #define Integral(x, p, notp) ((floor(x)*(p)) + max(fract(x)-(notp), 0.0))
  vec3 brick(vec3 pos)
 {
	const vec3 BrickColor = vec3(0.7, 0.2, 0.0);
	const vec3 MortarColor = vec3(0.5, 0.5, 0.5);
	const vec2 BrickSize = vec2(0.2);
	const vec2 BrickPct = vec2(0.85);
	const vec2 MortarPct = vec2(0.15);

	vec2 position, fw, useBrick;
	vec3 color;

	position = pos.xy / BrickSize;

	if (fract(position.y * 0.5) > 0.5)
		position.x += 0.5;

	// fwidth - return the sum of the absolute derivatives in x and y
	fw = fwidth(position);

	useBrick = (Integral(position + fw, BrickPct, MortarPct) - 
	            Integral(position, BrickPct, MortarPct)) / fw;

	return mix(MortarColor, BrickColor, useBrick.x * useBrick.y);
 }
#endif
 
void main()
{
    /*Todo : check the vector length before normalization (currently it is check before passed to the shader)*/
	vec3 normal = NormalToFrag;
	vec4 texBump;
	vec3 tangent, bitangent;
	if (Material.Has_Bump>0)
	{
		tangent = TangentToFrag.xyz;
        bitangent = cross(tangent.xyz,normal.xyz);
	
		texBump = texture(BumpMap, TexCoordToFrag);
		vec3 bumpNormal = (texBump.rgb*2.0 - 1.0);
		normal = bumpNormal.x*tangent + bumpNormal.y*bitangent + bumpNormal.z*normal;
		normal = normalize(normal);
	}

	
	// lighting vectors
	vec3 eyePos = vec3(0,0,0);
	//vec3 LightPos = LightPosToFrag.xyz / LightPosToFrag.w;
	//vec3 LightDir = normalize(LightDirToFrag.xyz);
	//vec3 LightToPx = normalize((PxPosToFrag - LightPosToFrag).xyz);
	vec3 LightPos = Light.Origin;
	vec3 LightDir = normalize(Light.Direction);
	vec3 LightToPx = normalize(PxPosToFrag.xyz/PxPosToFrag.w - LightPos);

	// color initialize
	vec3 ambient = vec3(0,0,0);
	vec3 diffuse = vec3(0,0,0);
	vec3 specular = vec3(0,0,0);
    
	// TODO tex color
	float lit = 1.0;
    float one_minus_ao = 1.0;
    
    if (Material.Receive_Light>0)
    {
        if (Light.Type==LIGHT_TYPE_AMBIENT)
        {
            if (Material.Const_Refl==0)
                ambient += Light.Color;
        }
        else if (Light.Type==LIGHT_TYPE_POINT)
        {
            if (ShadowControl.ShadowOnGlobal>0 && Light.ShadowOn>0 && Material.Receive_Shadow>0 && Material.Const_Refl==0)
            {
                float currentDepth = (length(PxPosinLightsView.xyz/PxPosinLightsView.w)-Light.NearClip)/(1.414*Light.FarClip-Light.NearClip);
                lit = ComputeShadow_Cubemap(normalize(PxPosinLightsView.xyz), currentDepth);
            }
            if (Material.Const_Refl==0)
                PointLight(eyePos, PxPosToFrag.xyz/PxPosToFrag.w, LightPos, normal, ambient, diffuse, specular);
            
        }
        else if (Light.Type==LIGHT_TYPE_DISTANT)
        {
            if (ShadowControl.ShadowOnGlobal>0 && Light.ShadowOn>0 && Material.Receive_Shadow>0 && Material.Const_Refl==0)
            {
                float currentDepth = depthLinear;
                lit = ComputeShadow_2D(ShadowMapTexCoord.st/ShadowMapTexCoord.w, currentDepth);
            }
            if (Material.Const_Refl==0)
                DirectionalLight(eyePos, PxPosToFrag.xyz/PxPosToFrag.w, LightDir, normal, ambient, diffuse, specular);
        }
        else if (Light.Type==LIGHT_TYPE_CONE)
        {
            if (ShadowControl.ShadowOnGlobal>0 && Light.ShadowOn>0 && Material.Receive_Shadow>0 && Material.Const_Refl==0)
            {
                float currentDepth = depthLinear;
                lit = ComputeShadow_2D(ShadowMapTexCoord.st/ShadowMapTexCoord.w, currentDepth);
            }
            if (Material.Const_Refl==0)
                ConeLight(eyePos, PxPosToFrag.xyz/PxPosToFrag.w, LightPos, LightDir, normal, ambient, diffuse, specular);
        }
        
        /*MJY 07/10/13 : ao is in effect if shadow is turned off*/
        if (AOControl.IsEnabled>0 && Material.Use_AO>0 /*&& Material.Is_Transparent==0*/ && Material.Const_Refl==0)
            //if (ShadowControl.ShadowOnGlobal>0 && AOControl.IsEnabled>0 && Material.Use_AO>0 && Material.Transparency>0.9999)
        {
            vec2 fragTexCoord=0.5*ProjPxPosToFrag.xy/ProjPxPosToFrag.w+0.5;
            one_minus_ao = texture(AmbientOcclusionMap, fragTexCoord).r;
        }
    }
	
    float intensity = 1.0/float(max(Light_Num,1));
  
    
    // use texture
	if (Material.Has_Texture>0)
	{
        vec4 tex_Color = texture(ColorTexture, TexCoordToFrag);
        if (Material.Receive_Light>0)
        {
            if (Material.Const_Refl==0)
            {
                FragColor.rgb = (ambient*Material.Ambient*one_minus_ao
                        + diffuse*Material.Diffuse*one_minus_ao
                        + specular*Material.Specular)*tex_Color.rgb*lit*Light.Factor/100.0;
                FragColor.a = (Material.Is_Transparent>0?Material.Transparency:1.0) * tex_Color.a;    // MJY 06/12/13
            }
            else
            {
                FragColor.rgb = vec3(intensity, intensity, intensity)*Material.Ambient*tex_Color.rgb*one_minus_ao*lit;
                FragColor.a = (Material.Is_Transparent>0?Material.Transparency:1.0) * tex_Color.a;    // MJY 06/12/13
            }
		}
        else
            FragColor = vec4(tex_Color.rgb, (Material.Is_Transparent>0?Material.Transparency:1.0) * tex_Color.a);
	}
    // use vertex color
    else if (Material.Has_Vertex_Color>0)
    {
        if (Material.Receive_Light>0)
        {
            if (Material.Const_Refl==0)
            {
                FragColor.rgb = (ambient*Material.Ambient*one_minus_ao
                        + diffuse*Material.Diffuse*one_minus_ao
                        + specular*Material.Specular)*ColorToFrag.rgb*lit*Light.Factor/100.0;
                FragColor.a = (Material.Is_Transparent>0?Material.Transparency:1.0) * ColorToFrag.a;    // MJY 06/12/13
            }
            else
            {
                FragColor.rgb = vec3(intensity, intensity, intensity)*Material.Ambient*ColorToFrag.rgb*one_minus_ao*lit;
                FragColor.a = (Material.Is_Transparent>0?Material.Transparency:1.0) * ColorToFrag.a;    // MJY 06/12/13
            }
		}
        else
            FragColor = vec4(ColorToFrag.rgb, (Material.Is_Transparent>0?Material.Transparency:1.0) * ColorToFrag.a);
    }
    else
	{
		vec3 tex_Color=vec3(1,1,1);

#if DO_PROC_TEX
        if (Material.Transparency>=0.95)
			tex_Color = granite(MCVertexToFrag*0.1);
		else if (Material.Transparency>=0.9)
			tex_Color = wood(MCVertexToFrag*0.1);
		else if (Material.Transparency>=0.8)
			tex_Color = marble(MCVertexToFrag*0.1);	
		else if (Material.Transparency>=0.7)
			tex_Color = cloud(MCVertexToFrag*0.1);
		else if (Material.Transparency>=0.6)
			tex_Color = strip(MCVertexToFrag*0.1);	
		else if (Material.Transparency>=0.5)
			tex_Color = ribbon(MCVertexToFrag*0.1);				
		else if (Material.Transparency>=0.4)
			tex_Color = brick(vec3(TexCoordToFrag,0));	
		else if (Material.Transparency>=0.3)
			tex_Color = alien(MCVertexToFrag*0.1);
		else if (Material.Transparency>=0.2)
			tex_Color = turbulence(MCVertexToFrag*0.1);	
		else if (Material.Transparency>=0.1)
			tex_Color = checker(vec3(TexCoordToFrag,0)*4);	
#endif
			
        if (Material.Receive_Light>0)
        {
            if (Material.Const_Refl==0)
            {
                FragColor.rgb = (ambient*Material.Ambient*one_minus_ao
                        + diffuse*Material.Diffuse*one_minus_ao
                        + specular*Material.Specular)*tex_Color*lit*Light.Factor/100.0;
#if DO_PROC_TEX
                FragColor.a = 1.0;//(Material.Is_Transparent>0?Material.Transparency:1.0);
#else
                FragColor.a = (Material.Is_Transparent>0?Material.Transparency:1.0);
#endif
            }
            else
            {
                FragColor.rgb = vec3(intensity, intensity, intensity)*Material.Ambient*one_minus_ao*lit;
#if DO_PROC_TEX
                FragColor.a = 1.0;//(Material.Is_Transparent>0?Material.Transparency:1.0);
#else
                FragColor.a = (Material.Is_Transparent>0?Material.Transparency:1.0);
#endif
            }
        }
		else
            FragColor = vec4(0.0, 0.0, 0.0, (Material.Is_Transparent>0?Material.Transparency:1.0));
	}


    if (FragColor.a<=ALPHA_THRES)
    {
        discard;
    }
}
